#include <gtk/gtk.h>
#include <glib/gstdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>

#include <main.h>

#include <bmp_playlist.h>
#include <bmp_metadata.h>
#include <g_object_get.h>
#include <util.h>
#include <uri.h>
#include <list.h>

#include <bmp_marshalers.h>

G_DEFINE_TYPE (BmpPlaylist, bmp_playlist, G_TYPE_OBJECT)

typedef enum {
        BMP_PLAYLIST_SIGNAL_LIST_DELETED,
        BMP_PLAYLIST_SIGNAL_LIST_ADDED,
        BMP_PLAYLIST_SIGNAL_LIST_RENAMED,

        BMP_PLAYLIST_SIGNAL_TRACKLIST_LIST_SORTED,
	BMP_PLAYLIST_SIGNAL_TRACKLIST_ITEMS_REMOVED,
	BMP_PLAYLIST_SIGNAL_TRACKLIST_ITEMS_ADDED,

	BMP_PLAYLIST_N_SIGNALS
} BmpPlaylistSignals;

static guint signals[BMP_PLAYLIST_N_SIGNALS] = { 0, 0, 0, 0, 0, 0};

typedef enum {
	BMP_PLAYLIST_PROP_0,

	BMP_PLAYLIST_N_PROPERTIES
} BmpPlaylistProperties;

struct _BmpPlaylistPrivate
{
        gboolean dispose_has_run;

	BmpListType *playlists;

	gchar *pre_current_name;
	gint pre_current_position;
};

/* Track List funcs */
gchar*
tracklist_column_data_func (BmpFileItem *item, gint column_idx)
{
    gchar *response = NULL;

#if 0
	while (item->locked) { while (gtk_events_pending()) gtk_main_iteration(); }
#endif

	switch (column_idx)
	{
		case 0:
		{
			if (item->metadata) {
				response = bmp_metadata_get_titlestring (item->metadata);
			} else {
				response = g_path_get_basename (item->uri);
			}
		}
		break;

		case 1:
		{

			gint ttime;
			gint minutes;
			gint seconds;
			const gchar *time;

			if (!item->metadata)
                            return g_strdup("");

			time = bmp_metadata_get_metadata_field(item->metadata, BMP_DATUM_TIME);

			if (time) {
			    ttime = strtol (time,NULL,10);
			}
			else
			    ttime = 0;

			minutes = ttime / 60;
			seconds = ttime % 60;

			response = g_strdup_printf("%d:%-2.2d",minutes,seconds);
		}
		break;
	}

	return response;
}

static void
tracklist_item_swap_func (GList *items, gint _idx1, gint _idx2)
{
	BmpFileItem *fi_a, *fi_b;

 	GList *a, *b;

	a = g_list_nth (items, _idx1);
	b = g_list_nth (items, _idx2);

	glist_swap(a,b);

	fi_a = a->data;
	fi_b = b->data;

	fi_a->dirty = TRUE;
	fi_b->dirty = TRUE;
}


void
tracklist_on_display_func (BmpFileItem *item)
{
	if (!item) return;
	if (item->metadata_attempted) return;
	if (item->metadata != NULL) return;

	while (item->locked) { while (gtk_events_pending()) gtk_main_iteration(); }

	item->metadata = bmp_metadata_get_from_uri (item->uri);
	item->metadata_attempted = TRUE;
	item->dirty = TRUE;
}

static gboolean
tracklist_get_dirty_func (BmpFileItem *item)
{
#if 0
	while (item->locked) { while (gtk_events_pending()) gtk_main_iteration(); }
#endif

	if (!item) return FALSE;
	return item->dirty;
}

static void
tracklist_set_dirty_func (BmpFileItem *item, gboolean dirty)
{
#if 0
	while (item->locked) { while (gtk_events_pending()) gtk_main_iteration(); }
#endif

	if (!item) return;
	item->dirty = dirty;
}

static gboolean
tracklist_get_rendered_func (BmpFileItem *item)
{
#if 0
	while (item->locked) { while (gtk_events_pending()) gtk_main_iteration(); }
#endif

	if (!item) return FALSE;
	return item->rendered;
}

static void
tracklist_set_rendered_func (BmpFileItem *item, gboolean rendered)
{
#if 0
	while (item->locked) { while (gtk_events_pending()) gtk_main_iteration(); }
#endif

	if (!item) return;
	item->rendered = rendered;
}


static gboolean
tracklist_get_selected_func (BmpFileItem *item)
{
#if 0
	while (item->locked) { while (gtk_events_pending()) gtk_main_iteration(); }
#endif

	return item->selected;
}

static void
tracklist_set_selected_func (BmpFileItem *item, gboolean selected)
{
#if 0
	while (item->locked) { while (gtk_events_pending()) gtk_main_iteration(); }
#endif

	item->selected = selected;
	item->dirty = TRUE;
}

static void
tracklist_item_destroy_func (BmpFileItem *item)
{
    g_free (item->uri);
    g_free (item);
}

/* Playlists List funcs */
static gchar*
playlists_column_data_func (BmpPlaylistItem *item, gint column_idx)
{
	if (!item) return NULL;

	return item->listname;
}

static gboolean
playlists_get_dirty_func (BmpPlaylistItem *item)
{
	if (!item) return FALSE;
	return item->dirty;
}

static void
playlists_set_dirty_func (BmpPlaylistItem *item, gboolean dirty)
{
	if (!item) return;
	item->dirty = dirty;
}

static gboolean
playlists_get_rendered_func (BmpPlaylistItem *item)
{
	if (!item) return FALSE;
	return item->dirty;
}

static void
playlists_set_rendered_func (BmpPlaylistItem *item, gboolean dirty)
{
	if (!item) return;
	item->dirty = dirty;
}


static gboolean
playlists_get_selected_func (BmpPlaylistItem *item)
{
	if (!item) return FALSE;

	return item->selected;
}

static void
playlists_set_selected_func (BmpPlaylistItem *item, gboolean selected)
{
	item->selected = selected;
	item->dirty = TRUE;
}

static void
playlists_item_destroy_func (BmpPlaylistItem *item)
{
	/*g_free (item->filename);
	g_free (item->listname);*/
}

static gint
playlists_compare_func (const BmpPlaylistItem *a, const BmpPlaylistItem *b)
{
	gchar *string_a;
	gchar *string_b;
	gint cmp_result;

	string_a = g_utf8_casefold (a->listname, strlen(a->listname));
	string_b = g_utf8_casefold (b->listname, strlen(b->listname));

	if ( (string_a != NULL) && (string_b == NULL)) return  1;
	if ( (string_a == NULL) && (string_b == NULL)) return  0;
	if ( (string_a == NULL) && (string_b != NULL)) return -1;

	cmp_result = g_utf8_collate (string_a, string_b);

	/*g_free (string_a);
	g_free (string_b);*/

	return cmp_result;
}

static gint
tracklist_compare_func (BmpFileItem *a, BmpFileItem *b, gpointer data)
{
	BmpMetadatumId tag_id;
	gchar *cmptext_a, *cmptext_b;
	gchar *string_a;
	gchar *string_b;
	gint cmp_result;
        gint val_a, val_b;

	tracklist_on_display_func (a);
	tracklist_on_display_func (b);

	if (! data)
	{
		cmptext_a = tracklist_column_data_func (a, 0);
		cmptext_b = tracklist_column_data_func (b, 0);

		if ( (cmptext_a != NULL) && (cmptext_b == NULL)) return  1;
		if ( (cmptext_a == NULL) && (cmptext_b == NULL)) return  0;
		if ( (cmptext_a == NULL) && (cmptext_b != NULL)) return -1;

	}
	else
	{
		tag_id = GPOINTER_TO_INT(data);

		cmptext_a = g_hash_table_lookup(a->metadata, bmp_metadata_get_id_static(tag_id));
		cmptext_b = g_hash_table_lookup(b->metadata, bmp_metadata_get_id_static(tag_id));

		if ( (cmptext_a != NULL) && (cmptext_b == NULL)) return  1;
		if ( (cmptext_a == NULL) && (cmptext_b == NULL)) return  0;
		if ( (cmptext_a == NULL) && (cmptext_b != NULL)) return -1;

		if ((tag_id == BMP_DATUM_TRACK) || (tag_id == BMP_DATUM_DATE))
		{
			if (! cmptext_a)
			    val_a = 0;
			else
			    val_a = strtol(cmptext_a,NULL,10);

			if (! cmptext_b)
			    val_b = 0;
			else
			    val_b = strtol(cmptext_b,NULL,10);

			return val_a - val_b;
		}
	}

	string_a = g_utf8_casefold (cmptext_a, strlen(cmptext_a));
	string_b = g_utf8_casefold (cmptext_b, strlen(cmptext_b));

	if ( (string_a != NULL) && (string_b == NULL)) return  1;
	if ( (string_a == NULL) && (string_b == NULL)) return  0;
	if ( (string_a == NULL) && (string_b != NULL)) return -1;

	cmp_result = g_utf8_collate (string_a, string_b);

	g_free (string_a);
	g_free (string_b);

	return cmp_result;
}


#if 0
static void
_print_listnames (BmpPlaylistItem *item, gpointer data)
{
	g_message ("LIST: %s", item->listname);
}
#endif

static void
playlists_sort_func (BmpPlaylist *self, gconstpointer data2)
{
	self->private->playlists->items =
			g_list_sort (self->private->playlists->items,
			(GCompareFunc)playlists_compare_func);

}

static void
tracklist_sort_func (BmpListType *list, gconstpointer data2)
{
	gpointer item;

	item = g_list_nth (list->items, list->current);
	list->items = g_list_sort_with_data (list->items, (GCompareDataFunc)tracklist_compare_func, (gpointer)data2);
	list->current = g_list_position (list->items, item);
}

static void
_playlists_clear_current (gpointer item, BmpPlaylist *self)
{
	BmpPlaylistItem *list_item = (BmpPlaylistItem*)item;

	list_item->list->current = -1;
}

void
bmp_playlist_sort_current_list (BmpPlaylist *self)
{
	BmpListType *list;
	GList *llink = NULL;
	gint _idx;

	_idx =
	    g_object_get_int(bmp_system_control, "current_tracklist");
	list =
	    bmp_playlist_tracklist_get_i (self, _idx);

	if (list->last_selected != -1)
	    llink =
		g_list_nth (list->items, list->last_selected);

	list->list_sort_func (list, NULL);

	if (llink)
	    list->last_selected =
		g_list_position (list->items, llink);

        g_signal_emit (G_OBJECT (self),
                   signals[BMP_PLAYLIST_SIGNAL_TRACKLIST_LIST_SORTED],
                   0, _idx);
}

void
bmp_playlist_sort_current_list_by_tag (BmpPlaylist *self, BmpMetadatumId tag_id)
{
	BmpListType *list;
	GList *llink = NULL;
	gint _idx;

	_idx =
	    g_object_get_int(bmp_system_control, "current_tracklist");
	list =
	    bmp_playlist_tracklist_get_i (self, _idx);

	if (list->last_selected != -1)
	    llink =
		g_list_nth (list->items, list->last_selected);

	list->list_sort_func (list, GINT_TO_POINTER(tag_id));

	if (llink)
	    list->last_selected =
		g_list_position (list->items, llink);

        g_signal_emit (G_OBJECT (self),
                   signals[BMP_PLAYLIST_SIGNAL_TRACKLIST_LIST_SORTED],
                   0, _idx);
}


void
bmp_playlist_clear_all_current (BmpPlaylist *self)
{
	g_return_if_fail (BMP_IS_PLAYLIST (self));

	g_list_foreach (self->private->playlists->items, (GFunc)_playlists_clear_current, self);
}

static void
load_playlist (BmpPlaylist *self, const gchar *filename)
{
	GList *uri_list;
	gchar *name, *aux;
	gchar **uri_strv;
	gint new_count = 0;
	gint list;

	g_return_if_fail (BMP_IS_PLAYLIST (self));

	aux = g_path_get_basename (filename);
	name = g_strndup (aux, strlen(aux) - 4);
	g_free (aux);

	list = bmp_playlist_tracklist_new (self, name);

	/* FIMXE: Use GURI for that */
	aux =
	    g_strconcat ("file://", filename, NULL);
	uri_list =
	    get_uris__m3u (aux);
	g_free (aux);

	uri_strv = glist_to_strv (uri_list);

	bmp_system_control_add_uri_list (bmp_system_control, uri_strv, list, -1, &new_count, NULL);

	g_list_foreach (uri_list, (GFunc) g_free, NULL);
	g_list_free (uri_list);

        g_signal_emit (G_OBJECT (self),
                   signals[BMP_PLAYLIST_SIGNAL_LIST_ADDED],
                   0, list);

#if 0
	if (!strcmp(name, self->private->pre_current_name) && (self->private->pre_current_position != -1)) {

		list->current = self->private->pre_current_position;
		self->private->pre_current_position = -1;
	}
	else
		list->current = -1;
#endif

	g_free(name);
}

#if 0
static void
playlist_removed_func (const gchar *filename, BmpPlaylist *self)
{
}

static void
playlist_added_func (const gchar *filename, BmpPlaylist *self)
{
	gchar *fqname;

	fqname = g_build_filename (bmp_configuration_key_fetch (bmp_configuration, "bmp", "playlists-path"),
				   filename, NULL);
	load_playlist (self, fqname);
	g_free (fqname);

}
#endif

static void
bmp_playlist_init (BmpPlaylist  *self)
{
	gchar *listname;
	gint _idx;

        self->private = g_new0(BmpPlaylistPrivate, 1);

        self->private->dispose_has_run = FALSE;

	self->private->pre_current_position = -1;
	self->private->pre_current_name = NULL;

	self->private->playlists = g_new0(BmpListType,1);
	self->private->playlists->name = NULL;
	self->private->playlists->column_data_func = (ColumnDataFunc)playlists_column_data_func;
	self->private->playlists->row_get_selected_func = (RowGetSelectedFunc)playlists_get_selected_func;
	self->private->playlists->row_set_selected_func = (RowSetSelectedFunc)playlists_set_selected_func;
	self->private->playlists->list_sort_func = (ListSortFunc)playlists_sort_func;
	self->private->playlists->on_display_func = NULL;
	self->private->playlists->get_dirty_func = (GetDirtyFunc)playlists_get_dirty_func;
	self->private->playlists->set_dirty_func = (SetDirtyFunc)playlists_set_dirty_func;
	self->private->playlists->item_destroy_func = (GDestroyNotify)playlists_item_destroy_func;
	self->private->playlists->set_rendered_func = (SetRenderedFunc)playlists_set_rendered_func;
	self->private->playlists->get_rendered_func = (GetRenderedFunc)playlists_get_rendered_func;
	self->private->playlists->items = NULL;
	self->private->playlists->current = -1;
	self->private->playlists->last_selected = -1;
	self->private->playlists->multiple_selected = FALSE;

#if 0
	fam_monitor_watch_new (bmp_fam_monitor, "playlists",
				bmp_configuration_key_fetch(bmp_configuration, "bmp", "playlists-path"),
				playlist_added_func,
				playlist_removed_func,
				self,
				NULL);
#endif

    listname =
	bmp_configuration_key_fetch(bmp_configuration,"bmp","current-playlist");
    _idx =
	bmp_configuration_key_fetch_int(bmp_configuration,"bmp","current-playlist-position");

    if (! listname) listname = g_strdup("");

    bmp_playlist_set_pre_current_name (self, listname);
    bmp_playlist_set_pre_current_position (self, _idx);

    g_free (listname);

    g_message("%s:\t\tINITIALIZED", __FILE__);
}

static void
bmp_playlist_set_property (GObject      *object,
                        guint         property_id,
                        const GValue *value,
                        GParamSpec   *pspec)
{
        switch (property_id) {

        default:
                /* We don't have any other property... */
                g_assert (FALSE);
                break;
        }
}

static void
bmp_playlist_get_property (GObject      *object,
                        guint         property_id,
                        GValue       *value,
                        GParamSpec   *pspec)
{
        switch (property_id) {

        default:
                /* We don't have any other property... */
                g_assert (FALSE);
                break;
        }
}

static GObject *
bmp_playlist_constructor (GType                  type,
                       guint                  n_construct_properties,
                       GObjectConstructParam *construct_properties)
{
        GObject *obj;

        {
                /* Invoke parent constructor. */
                BmpPlaylistClass *klass;
                GObjectClass *parent_class;
                klass = BMP_PLAYLIST_CLASS (g_type_class_peek (BMP_TYPE_PLAYLIST));
                parent_class = G_OBJECT_CLASS (g_type_class_peek_parent (klass));
                obj = parent_class->constructor (type,
                                                 n_construct_properties,
                                                 construct_properties);
        }

        /* do stuff. */

        return obj;
}

static void
bmp_playlist_dispose (GObject *object)
{
    BmpPlaylist *self = (BmpPlaylist *) object;

    if (self->private->dispose_has_run)
        return;

    self->private->dispose_has_run = TRUE;

    bmp_playlist_save_lists (self);

    g_message("%s:\t\tSTOPPED", __FILE__);
}

static void
bmp_playlist_finalize (GObject *object)
{
    /*
     * Here, complete object destruction.
     * You might not need to do much...
     */

#if 0
    BmpPlaylist *self = (BmpPlaylist *) gobject;
    g_free (self->private);
#endif
}

BmpPlaylist*
bmp_playlist_new (void)
{
  BmpPlaylist *list;
  list = g_object_new (bmp_playlist_get_type (), NULL);

  return list;
}

static void
bmp_playlist_class_init (BmpPlaylistClass *g_class)

{
    GObjectClass *gobject_class = G_OBJECT_CLASS (g_class);

    gobject_class->set_property = bmp_playlist_set_property;
    gobject_class->get_property = bmp_playlist_get_property;
    gobject_class->dispose = bmp_playlist_dispose;
    gobject_class->finalize = bmp_playlist_finalize;
    gobject_class->constructor = bmp_playlist_constructor;

    signals[BMP_PLAYLIST_SIGNAL_LIST_DELETED] =
        g_signal_new ("list-deleted",
                      G_OBJECT_CLASS_TYPE (gobject_class),
                      G_SIGNAL_RUN_FIRST,
                      G_STRUCT_OFFSET (BmpPlaylistClass, list_deleted),
                      NULL, NULL, g_cclosure_marshal_VOID__INT, G_TYPE_NONE,
                      1, G_TYPE_INT);

    signals[BMP_PLAYLIST_SIGNAL_LIST_ADDED] =
        g_signal_new ("list-added",
                      G_OBJECT_CLASS_TYPE (gobject_class),
                      G_SIGNAL_RUN_FIRST,
                      G_STRUCT_OFFSET (BmpPlaylistClass, list_added),
                      NULL, NULL, g_cclosure_marshal_VOID__INT, G_TYPE_NONE,
                      1, G_TYPE_INT);

    signals[BMP_PLAYLIST_SIGNAL_LIST_RENAMED] =
        g_signal_new ("list-renamed",
                      G_OBJECT_CLASS_TYPE (gobject_class),
                      G_SIGNAL_RUN_FIRST,
                      G_STRUCT_OFFSET (BmpPlaylistClass, list_renamed),
                      NULL, NULL, g_cclosure_marshal_VOID__INT, G_TYPE_NONE,
                      1, G_TYPE_INT);

    signals[BMP_PLAYLIST_SIGNAL_TRACKLIST_LIST_SORTED] =
        g_signal_new ("tracklist-list-sorted",
                      G_OBJECT_CLASS_TYPE (gobject_class),
                      G_SIGNAL_RUN_FIRST,
                      G_STRUCT_OFFSET (BmpPlaylistClass, tracklist_list_sorted),
                      NULL, NULL, g_cclosure_marshal_VOID__INT, G_TYPE_NONE,
                      1, G_TYPE_INT);

    signals[BMP_PLAYLIST_SIGNAL_TRACKLIST_ITEMS_REMOVED] =
        g_signal_new ("tracklist-items-removed",
                      G_OBJECT_CLASS_TYPE (gobject_class),
                      G_SIGNAL_RUN_FIRST,
                      G_STRUCT_OFFSET (BmpPlaylistClass, tracklist_items_removed),
                      NULL, NULL, g_cclosure_user_marshal_VOID__INT_POINTER_INT, G_TYPE_NONE,
                      3, G_TYPE_INT, G_TYPE_POINTER, G_TYPE_INT);

    signals[BMP_PLAYLIST_SIGNAL_TRACKLIST_ITEMS_ADDED] =
        g_signal_new ("tracklist-items-added",
                      G_OBJECT_CLASS_TYPE (gobject_class),
                      G_SIGNAL_RUN_FIRST,
                      G_STRUCT_OFFSET (BmpPlaylistClass, tracklist_items_added),
                      NULL, NULL, g_cclosure_user_marshal_VOID__INT_POINTER_INT, G_TYPE_NONE,
                      3, G_TYPE_INT, G_TYPE_POINTER, G_TYPE_INT);
}

BmpListType*
bmp_playlist_get_playlists (BmpPlaylist *self)
{
	return self->private->playlists;
}

const gchar*
bmp_playlist_get_nth_tracklist_name (BmpPlaylist *self, gint _idx)
{
	BmpPlaylistItem *item;

	g_return_val_if_fail (BMP_IS_PLAYLIST (self), NULL);

	if (! self->private->playlists->items)
	    return NULL;
	if (g_list_length (self->private->playlists->items) == 0)
	    return NULL;
	if (_idx > (g_list_length (self->private->playlists->items)-1))
	    return NULL;

	item = LIST_GET_NTH_ITEM(self->private->playlists, _idx);

	return item->listname;
}

void
bmp_playlist_set_pre_current_position (BmpPlaylist *self, gint _idx)
{
	g_return_if_fail (BMP_IS_PLAYLIST (self));

	self->private->pre_current_position = _idx;
}

void
bmp_playlist_set_pre_current_name (BmpPlaylist *self, const gchar *name)
{
	g_return_if_fail (BMP_IS_PLAYLIST (self));

	self->private->pre_current_name = g_strdup (name);
}

const gchar*
bmp_playlist_get_current_tracklist_name (BmpPlaylist *self)
{
	BmpPlaylistItem *item;
	gint _idx;

	g_return_val_if_fail (BMP_IS_PLAYLIST (self), NULL);

	if (! self->private->playlists)
	    return NULL;
	if (! self->private->playlists->items)
	    return NULL;
	if (g_list_length(self->private->playlists->items) == 0)
	    return NULL;

	_idx = g_object_get_int (bmp_system_control, "current_tracklist");

	if (_idx < 0)
	    return NULL;

	item = LIST_GET_NTH_ITEM(self->private->playlists, _idx);

	return item->listname;
}

gint
bmp_playlist_get_tracklist_by_name (BmpPlaylist *self, const gchar *name)
{
	BmpPlaylistItem *item;
	GList *iter_list;
	gint ctr = 0;

	g_return_val_if_fail (BMP_IS_PLAYLIST (self), -1);

	iter_list = self->private->playlists->items;

	while (iter_list)
	{
		item = iter_list->data;
		if (!strcmp(name, item->listname)) break;
		ctr++;

		iter_list = g_list_next (iter_list);
	}

	return ctr;
}

void
bmp_playlist_tracklist_shuffle (BmpPlaylist *self, gint _idx)
{
	BmpListType *item;

	g_return_if_fail (BMP_IS_PLAYLIST (self));

	item = bmp_playlist_tracklist_get_i (self, _idx);

	if (!item)
		return;

	if (!item->items)
		return;

	item->items_shuffle = glist_make_shuffle_index (item->items);

	item->current = 0;
	item->current_linear = 0;
}

static gboolean
playlist_rename (BmpPlaylist *self, const gchar *listname, const gchar *newname)
{
	BmpListType *tracklist;
	BmpPlaylistItem *playlist_item;
	GList *iter_list;
	gchar *filename, *aux1;

	g_return_val_if_fail (BMP_IS_PLAYLIST (self), FALSE);

	tracklist = bmp_playlist_tracklist_get (self, newname);
	if (tracklist) return FALSE;

	iter_list = g_list_first(self->private->playlists->items);

	while (iter_list)
	{
		playlist_item = iter_list->data;

		if (!strcmp(playlist_item->listname, listname))
		{
			aux1 = g_strconcat (newname, ".m3u", NULL);
			filename = g_build_filename (bmp_paths[BMP_PATH_PLAYLISTS], aux1, NULL);

			g_rename (playlist_item->filename, filename);

			g_free (aux1);
			g_free (filename);
			break;
		}
		iter_list = g_list_next (iter_list);
	}

	return TRUE;
}

gboolean
bmp_playlist_rename_playlist (BmpPlaylist *self, gint _idx, const gchar *name)
{
	BmpPlaylistItem *item;
	GList *llink_current = NULL, *llink_active = NULL, *llink = NULL;
	gchar *name_old, *aux1, *aux2;

	llink = g_list_nth (self->private->playlists->items, self->private->playlists->current);
	llink_current = LIST_GET_NTH_ITEM_GLIST(self->private->playlists, g_object_get_int (bmp_system_control, "current_tracklist"));
	llink_active  = LIST_GET_NTH_ITEM_GLIST(self->private->playlists, g_object_get_int (bmp_system_control, "active_tracklist"));

	item = LIST_GET_NTH_ITEM(self->private->playlists, _idx);
	name_old = g_strdup (item->listname);
	if (! playlist_rename (self, name_old, name)) return FALSE;

	aux1 = g_strconcat(name, ".m3u", NULL);
	aux2 = g_build_filename (bmp_paths[BMP_PATH_PLAYLISTS], aux1, NULL);
	g_free (item->listname);
	item->listname = g_strdup(name);
	g_free (item->filename);
	item->filename = g_strdup(aux2);

	self->private->playlists->list_sort_func (self, NULL);

	self->private->playlists->current = g_list_position(self->private->playlists->items, llink);
	g_object_set (bmp_system_control, "current_tracklist", (gint)g_list_position(self->private->playlists->items, (GList*)llink_current), NULL);
	g_object_set (bmp_system_control, "active_tracklist",  (gint)g_list_position(self->private->playlists->items, (GList*)llink_active), NULL);

	g_free (aux1);
	g_free (aux2);
	g_free (name_old);

        g_signal_emit (G_OBJECT (self),
                   signals[BMP_PLAYLIST_SIGNAL_LIST_RENAMED],
                   0, _idx);

	return TRUE;
}

void
bmp_playlist_remove_all_playlists (BmpPlaylist *self)
{
	BmpPlaylistItem *item;
	GList *iter_list;

	iter_list = self->private->playlists->items;

	while (iter_list)
	{
		item = iter_list->data;
		g_unlink (item->filename);
		iter_list = g_list_next (iter_list);
	}

	self->private->playlists->current = -1;
	self->private->playlists->items = NULL;

	/* FIXME: This is kinda wrong */
        g_signal_emit (G_OBJECT (self),
                   signals[BMP_PLAYLIST_SIGNAL_LIST_DELETED],
                   0);
}

void
bmp_playlist_remove_playlist (BmpPlaylist *self, gint _idx)
{
	BmpPlaylistItem *item;
	GList *llink_current = NULL, *llink_active = NULL, *llink = NULL;

	item = g_list_nth(self->private->playlists->items, _idx)->data;
	g_unlink (item->filename);

	llink_current = g_list_nth(self->private->playlists->items, g_object_get_int (bmp_system_control, "current_tracklist"));
	llink_active  = g_list_nth(self->private->playlists->items, g_object_get_int (bmp_system_control, "active_tracklist"));
	llink = g_list_nth (self->private->playlists->items, self->private->playlists->current);

	LIST_REMOVE_ITEM(self->private->playlists, _idx);

	self->private->playlists->current = g_list_position (self->private->playlists->items, llink);
	g_object_set (bmp_system_control, "current_tracklist", (gint)g_list_position(self->private->playlists->items, llink_current), NULL);
	g_object_set (bmp_system_control, "active_tracklist",  (gint)g_list_position(self->private->playlists->items, llink_active), NULL);

        g_signal_emit (G_OBJECT (self),
                   signals[BMP_PLAYLIST_SIGNAL_LIST_DELETED],
                   0, _idx);
}

#if 0
static void
playlist_remove_playlist_by_name (BmpPlaylist *self, const gchar *name)
{
}
#endif

static gint
insert_playlist_item (BmpPlaylist *self, const gchar *name, BmpListType *list)
{
	BmpPlaylistItem *playlist_item = NULL;
	GList *llink;
	gchar *basename;
	gchar *filename;

	llink = g_list_nth (self->private->playlists->items, self->private->playlists->current);

	basename = g_strconcat (name, ".m3u", NULL);
	filename = g_build_filename (bmp_paths[BMP_PATH_PLAYLISTS], basename, NULL);
	g_free (basename);

	playlist_item = g_new0(BmpPlaylistItem,1);
	playlist_item->filename = filename;
	playlist_item->listname = g_strdup(name);
	playlist_item->selected = FALSE;
	playlist_item->dirty = FALSE;
	playlist_item->rendered = FALSE;
	playlist_item->list = list;
	LIST_ADD_ITEM(self->private->playlists, playlist_item);

	self->private->playlists->list_sort_func (self, NULL);
	self->private->playlists->current = g_list_position (self->private->playlists->items, llink);

	return g_list_position (self->private->playlists->items, g_list_find (self->private->playlists->items, playlist_item));
}

static BmpFileItem*
file_item_new (const gchar *uri) {

    BmpFileItem *file_item = NULL;

    g_return_val_if_fail (uri != NULL, NULL);

    file_item = g_new0(BmpFileItem,1);

    file_item->uri = g_strdup(uri);
    file_item->metadata = NULL;
    file_item->metadata_attempted = FALSE;
    file_item->selected = FALSE;
    file_item->dirty = FALSE;
    file_item->rendered = FALSE;

    return file_item;
}

static int
bmp_playlist_insert_tracklist_item_position_R (BmpPlaylist *self, gint tracklist, const gchar *uri, gint _idx)
{
	BmpListType *list;
	BmpFileItem *file_item = NULL;
	GList *items = NULL;
	GError *error = NULL;
	GURI *g_uri;

	list =
	    bmp_playlist_tracklist_get_i (self, tracklist);

	g_uri =
	    gnet_uri_new (uri);

	if (! g_uri)
	    return -1;

	gnet_uri_unescape (g_uri);

	if (str_has_suffix_nocase(uri, ".pls") || str_has_suffix_nocase(uri, ".m3u")) {

		GList *uri_list = NULL;

		uri_list =
		    get_uris__playlist (gnet_uri_get_string(g_uri));

		while (uri_list) {
		    file_item =
			file_item_new (uri_list->data);
		    LIST_ADD_ITEM_POSITION(bmp_playlist_tracklist_get_i (self, tracklist), file_item, _idx++);
		    uri_list =
			g_list_next (uri_list);
		}

		g_list_foreach (uri_list, (GFunc) g_free, NULL);
		g_list_free (uri_list);

	} else if ( (g_file_test(g_uri->path, G_FILE_TEST_IS_REGULAR)) || !strcmp("cdda",g_uri->scheme)){
			file_item = file_item_new (gnet_uri_get_string(g_uri));
			LIST_ADD_ITEM_POSITION(list, file_item, _idx++);
	} else if (g_file_test(g_uri->path, G_FILE_TEST_IS_DIR)) {

		dir_foreach(g_uri->path, (DirForeachFunc)collect_path,
                 	    &items, &error);

		while (items) {
			GURI *g_uri;
			gchar *new_uri = g_filename_to_uri (items->data, NULL, NULL);

			g_uri = gnet_uri_new (new_uri);
			gnet_uri_unescape (g_uri);
			g_free (new_uri);
			new_uri = gnet_uri_get_string (g_uri);

			/* Do not need to increase _idx here, as if its a file,
			 * above code will handle that ... */
			_idx = bmp_playlist_insert_tracklist_item_position_R (self, tracklist, new_uri, _idx);
			items = g_list_next (items);

			gnet_uri_delete (g_uri);
		}

		g_list_foreach (g_list_first(items), (GFunc) g_free, NULL);
        }

    if (g_object_get_boolean(bmp_system_control,"shuffle"))
	    bmp_playlist_tracklist_shuffle (self, tracklist);

    gnet_uri_delete (g_uri);

    return _idx;
}

gint
bmp_playlist_insert_tracklist_items_position (BmpPlaylist *self, gint tracklist, const gchar **uri_list, gint _idx)
{
    gint n = 0, n_added_rows = 0, _idx_original = _idx;
    GArray *added_rows;

    while (uri_list[n])
      {
        _idx = bmp_playlist_insert_tracklist_item_position_R (self, tracklist, uri_list[n], _idx);
	n++;
      }
    
    added_rows = g_array_new (FALSE, FALSE, sizeof(gint)); 

    for (n = 0; n < (_idx - _idx_original); n++)
      {
	gint val =
	    n + _idx_original;
	g_array_append_val (added_rows, val); 
      }

    n_added_rows = (_idx - _idx_original);

    g_signal_emit (G_OBJECT (self),
                   signals[BMP_PLAYLIST_SIGNAL_TRACKLIST_ITEMS_ADDED],
                   0, tracklist, added_rows, n_added_rows);

    return _idx;
}

gint
bmp_playlist_insert_tracklist_items (BmpPlaylist *self, gint tracklist, const gchar **uri_list)
{
    BmpListType *list;
    gint _idx;

    list = bmp_playlist_tracklist_get_i (self, tracklist);

    if (!list)
    {
        g_message ("%s: NO LIST! returning.", G_STRLOC);
        return -1;
    }

    _idx = 
	g_list_length (list->items);
    _idx = 
	bmp_playlist_insert_tracklist_items_position (self, tracklist, uri_list, _idx);

   return _idx;
}

void
bmp_playlist_tracklist_remove_all (BmpPlaylist *self, gint tracklist)
{
	BmpListType *list;
	GArray *removed_rows;
	gint n_removed_rows = -1;

	removed_rows = g_array_new (FALSE,FALSE,sizeof(gint));

	list = bmp_playlist_tracklist_get_i (self, tracklist);
	g_list_foreach (list->items, (GFunc) tracklist_item_destroy_func, NULL);
	list->items = NULL;
	list->current = 0;
	list->current_linear = 0;
	list->last_selected = LAST_SELECTED_NONE;
	list->adjustment = 0.0;

        g_signal_emit (G_OBJECT (self),
                   signals[BMP_PLAYLIST_SIGNAL_TRACKLIST_ITEMS_REMOVED],
                   0, tracklist, removed_rows, n_removed_rows);

	/* NULL means: All. heh. */
}

void
bmp_playlist_tracklist_remove_selected (BmpPlaylist *self, gint tracklist)
{
	GList *new_items;
	BmpListType *list;
	BmpFileItem *file_item;
	gint n, new_current, new_count, n_removed_rows = 0;
	GArray *removed_rows;

	list = 
	    bmp_playlist_tracklist_get_i (self, tracklist);
	new_items = 
	    NULL;
	new_count = 
	    0;
	new_current = 
	    -1;
	removed_rows =
	    g_array_new (FALSE, FALSE, sizeof(gint));

	for (n = 0; n < g_list_length(list->items); n++)
	  {
		file_item = LIST_GET_NTH_ITEM (list, n);

		/* FIXME: is it really ok to access privately? by doing this be basically say, the function pointers
			  in the list struct are only for the chromalist widget.. */
		if (!file_item->selected)
		  {
			new_items = g_list_append (new_items, file_item);
			if (n == list->current)
				new_current = new_count;
			new_count++;
		  }
		else
                  {
		    g_array_append_val (removed_rows, n);
		    n_removed_rows++;
		  }
	   }
	list->items = new_items;
	list->current = new_current;

	if ((list->last_selected > g_list_length(list->items)-1) && new_items)
	    list->last_selected = g_list_length(list->items)-1;
	else if (! new_items)
	    list->last_selected = LAST_SELECTED_NONE;

	if (list->last_selected != LAST_SELECTED_NONE)
	{
	    file_item = g_list_nth(list->items, list->last_selected)->data;
	    list->row_set_selected_func ((gpointer)file_item, TRUE);
	}
	else
	{
	    list->last_selected = LAST_SELECTED_NONE;
	}

        g_signal_emit (G_OBJECT (self),
                   signals[BMP_PLAYLIST_SIGNAL_TRACKLIST_ITEMS_REMOVED],
                   0, tracklist, removed_rows, n_removed_rows);

	/* FIXME: Free old items */
}

void
bmp_playlist_tracklist_remove_unselected (BmpPlaylist *self, gint tracklist)
{
	GList *new_items;
	BmpListType *list;
	BmpFileItem *file_item;
	gint n, new_count, new_current, n_removed_rows = 0;
	GArray *removed_rows;

	list =
	    bmp_playlist_tracklist_get_i (self, tracklist);
	new_items = 
	    NULL;
	new_count = 
	    0;
	new_current = 
	    -1;
	removed_rows =
	    g_array_new (FALSE, FALSE, sizeof(gint));

	for (n = 0; n < g_list_length(list->items); n++)
	  {
		file_item = LIST_GET_NTH_ITEM (list, n);

		/* fixme: is it really ok to access privately? by doing this be basically say, the function pointers
			  in the list struct are only for the chromalist widget.. */
		if (file_item->selected)
		  {
			new_items = g_list_append (new_items, file_item);
			if (n == list->current)
				new_current = new_count;
			new_count++;
		  }
		else
	          {
			g_array_append_val (removed_rows, n);
			n_removed_rows++;
		  }
	  }

	list->items = new_items;
	list->current = new_current;
	list->last_selected = 0;

        g_signal_emit (G_OBJECT (self),
                   signals[BMP_PLAYLIST_SIGNAL_TRACKLIST_ITEMS_REMOVED],
                   0, tracklist, removed_rows, n_removed_rows);

	/* FIXME: Free old items */
}


static gint
tracklist_find_func (gconstpointer a, gconstpointer b)
{
    BmpPlaylistItem *playlist_item = (BmpPlaylistItem*)a;

    if (!a) return 1;

    if (!strcmp(playlist_item->listname, (gchar*)b)) return 0;
    return 1;
}

BmpListType*
bmp_playlist_tracklist_get_i (BmpPlaylist *self, gint _idx)
{
	BmpPlaylistItem *playlist_item;
	GList *llink;

	llink = g_list_nth (self->private->playlists->items, _idx);
	if (! llink)
	    return NULL;

	playlist_item = llink->data;
	if (! playlist_item)
	{
	    g_message("%s: No link data; returning.", G_STRLOC);
	    return NULL;
	}

	return playlist_item->list;
}


BmpListType*
bmp_playlist_tracklist_get (BmpPlaylist *self, const gchar *name)
{
	BmpPlaylistItem *playlist_item;
	GList *llink;

	if (!name)
	    return NULL;

	llink = g_list_find_custom (self->private->playlists->items, name, tracklist_find_func);
	if (!llink)
	    return NULL;

	playlist_item = llink->data;
	if (!playlist_item)
	    return NULL;

	return playlist_item->list;
}

gint
bmp_playlist_tracklist_new (BmpPlaylist *self, const gchar *name)
{
	BmpListType *list;
	gint rv;

	list = bmp_playlist_tracklist_get (self, name);
	if (list)
	  {
	    g_message ("%s: List exists; returning", G_STRFUNC);
	    return -1;
          }

	list = g_new0(BmpListType,1);

	list->name = g_strdup (name);
	list->column_data_func = (ColumnDataFunc)tracklist_column_data_func;
	list->row_get_selected_func = (RowGetSelectedFunc)tracklist_get_selected_func;
	list->row_set_selected_func = (RowSetSelectedFunc)tracklist_set_selected_func;
	list->on_display_func = (OnDisplayFunc)tracklist_on_display_func;
	list->item_swap_func = (ItemSwapFunc)tracklist_item_swap_func;
	list->item_destroy_func = (GDestroyNotify)tracklist_item_destroy_func;
	list->list_sort_func = (ListSortFunc)tracklist_sort_func;
	list->set_dirty_func = (SetDirtyFunc)tracklist_set_dirty_func;
	list->get_dirty_func = (GetDirtyFunc)tracklist_get_dirty_func;
	list->set_rendered_func = (SetRenderedFunc)tracklist_set_rendered_func;
	list->get_rendered_func = (GetRenderedFunc)tracklist_get_rendered_func;
	list->items = NULL;
	list->items_shuffle = NULL;
	list->current = 0;
	list->multiple_selected = FALSE;
	list->last_selected = -1;
	list->adjustment = 0.0;

	rv = insert_playlist_item (self, name, list);

	if (g_object_get_boolean (bmp_system_control, "shuffle"))
          {
	    bmp_playlist_tracklist_shuffle (self, rv);
	  }

        g_signal_emit (G_OBJECT (self),
                   signals[BMP_PLAYLIST_SIGNAL_LIST_ADDED],
                   0, rv);

	return rv;
}

static void
save_playlist (BmpPlaylistItem *playlist_item, BmpPlaylist *self)
{
	BmpFileItem *item;
	GList *iter_list;
	GError *error = NULL;
	gchar *buffer = NULL, *aux;

	iter_list = playlist_item->list->items;

	while (iter_list)
	{
		item = iter_list->data;
		if (buffer!=NULL)
		{
			aux = g_strconcat (buffer, "\n", item->uri, NULL);
			g_free (buffer);
			buffer = g_strdup(aux);
			g_free (aux);
		}
		else
		{
			buffer = g_strconcat (item->uri, NULL);
		}

		iter_list = g_list_next(iter_list);
	}

	if ( (! buffer) || (! strlen(buffer))) {
		    buffer = g_strdup("\n");
	}

	g_file_set_contents (playlist_item->filename, buffer, -1, &error);

	g_free (buffer);

	if (error) {
		g_message("SAVE playlist [%s] ERROR: %s", playlist_item->filename, error->message);
	}
}

void
bmp_playlist_save_lists (BmpPlaylist *self)
{
	const gchar *name;

	g_return_if_fail (BMP_IS_PLAYLIST (self));

	if (!bmp_configuration)
	    return;

	if (g_list_length(self->private->playlists->items) == 0) return;
	g_list_foreach (self->private->playlists->items, (GFunc) save_playlist, self);

	/* FIXME: This doesn't belong here, it's an UI issue */
	name = bmp_playlist_get_current_tracklist_name (self);
	bmp_configuration_key_set_string (bmp_configuration,
		"bmp",
		"current-playlist",
		name);
#if 0
	list =
	    bmp_playlist_tracklist_get (self, name);
	if (list)
	    bmp_configuration_key_set_int (bmp_configuration,
		"bmp",
		"current-playlist-position",
		list->current);
#endif
}

void
bmp_playlist_load_playlists (BmpPlaylist *self)
{
	GDir *dir;
	GError *error = NULL;
	gchar *fqname, *listname;
	const gchar* filename;
	gint current_list = 0;
	gint list;

	g_return_if_fail (BMP_IS_PLAYLIST (self));

	dir = g_dir_open (bmp_configuration_key_fetch (bmp_configuration, "bmp", "playlists-path"), 0, &error);

	if (error)
	{
		g_message("LOAD PLAYLISTS ERORR: %s", error->message);
		return;
	}

	while (42)
	{
		filename = g_dir_read_name (dir);
		if (filename == NULL) break;
		if (str_has_suffix_nocase(filename, "m3u"))
		{
			fqname = g_build_filename (bmp_paths[BMP_PATH_PLAYLISTS], filename, NULL);
			load_playlist (self, fqname);
			g_free (fqname);
			current_list++;
		}
	}
	g_dir_close (dir);

	if (! current_list) {
		listname =
		    g_strdup ("Default");
		list =
		    bmp_playlist_tracklist_new (self, listname);

		if (list != -1) {
			bmp_playlist_set_pre_current_name (self, listname);
			bmp_playlist_set_pre_current_position (self, -1);
			bmp_configuration_key_set_string (bmp_configuration, "bmp", "current-playlist", "Default");
		} else
			g_free (listname);
	}

	self->private->playlists->list_sort_func (self, NULL);
}

gint
bmp_playlist_get_current_track_idx (BmpPlaylist *self)
{
	BmpListType *list;
	GList *llink;
	gpointer data;

	if (! self->private->playlists->items)
	    return -1;
	llink = g_list_nth (self->private->playlists->items, g_object_get_int (bmp_system_control, "active_tracklist"));
	if (! llink)
	    return -1;

	data  = llink->data;
	if (! data)
	    return -1;
	list  = ((BmpPlaylistItem*)data)->list;
	if (! list)
	    return -1;

	if (! list->items)
	    return -1;

	if (list->current < 0)
	    return -1;

	return list->current;
}

gchar*
bmp_playlist_get_current_uri (BmpPlaylist *self)
{
	BmpListType *list;
	BmpFileItem *item;
	GList *llink;
	gpointer data;
	gchar *uri;

	if (! self->private->playlists->items)
	    return NULL;
	llink = g_list_nth (self->private->playlists->items, g_object_get_int (bmp_system_control, "active_tracklist"));
	if (! llink)
	    return NULL;
	data  = llink->data;
	if (! data)
	    return NULL;
	list  = ((BmpPlaylistItem*)data)->list;
	if (! list)
	    return NULL;

	if (! list->items)
	    return NULL;

	if (list->current < 0)
	    return NULL;

	item = g_list_nth(list->items, list->current)->data;
	if (! item) return NULL;

	tracklist_on_display_func (item);

	uri = g_hash_table_lookup (item->metadata, bmp_metadata_get_id_static(BMP_DATUM_LOCATION));

	return uri;
}

gchar*
bmp_playlist_get_current_title (BmpPlaylist *self)
{
	BmpListType *list;
	BmpFileItem *item;
	GList *llink;
	gpointer data;
	gchar *title;

	if (! self->private->playlists->items)
	    return NULL;
	llink = g_list_nth (self->private->playlists->items, g_object_get_int (bmp_system_control, "active_tracklist"));
	if (! llink)
	    return NULL;
	data  = llink->data;
	if (! data)
	    return NULL;
	list  = ((BmpPlaylistItem*)data)->list;
	if (! list)
	    return NULL;

	if (! list->items)
	    return NULL;

	if (list->current < 0)
	    return NULL;

	item = g_list_nth(list->items, list->current)->data;
	if (! item)
	    return NULL;

	tracklist_on_display_func (item);

	title = bmp_metadata_get_titlestring (item->metadata);

	return title;
}
