/* -*- Mode: C; tab-width: 8; indent-tabs-mode: nil; c-basic-offset: 8 -*- */

#ifndef FAM_MONITOR_H
#define FAM_MONITOR_H

/*
 * Potentially, include other headers on which this header depends.
 */
#define PATH_MAX 4096

#include <glib-object.h>
#include <fam.h>

#define FAM_TYPE_MONITOR		  (fam_monitor_get_type ())
#define FAM_MONITOR(obj)		  (G_TYPE_CHECK_INSTANCE_CAST ((obj), FAM_TYPE_MONITOR, FamMonitor))
#define FAM_MONITOR_CLASS(klass)	  (G_TYPE_CHECK_CLASS_CAST ((klass), FAM_TYPE_MONITOR, FamMonitorClass))
#define FAM_IS_MONITOR(obj)	  (GTK_CHECK_TYPE ((obj), FAM_TYPE_MONITOR))
#define FAM_IS_MONITOR_CLASS(klass) (G_TYPE_CHECK_CLASS_TYPE ((klass), FAM_TYPE_MONITOR))
#define FAM_MONITOR_GET_CLASS(obj)  (G_TYPE_INSTANCE_GET_CLASS ((obj), FAM_TYPE_MONITOR, FamMonitorClass))

typedef struct _FamMonitor FamMonitor;
typedef struct _FamMonitorClass FamMonitorClass;
typedef struct _FamMonitorPrivate FamMonitorPrivate;

struct _FamMonitor {
	GObject parent;

        FamMonitorPrivate *private;
};

struct _FamMonitorClass {
	GObjectClass parent;
};

typedef void    (*FamAddedFunc)           (gchar *filename, gpointer user_data);
typedef void    (*FamRemovedFunc)         (gchar *filename, gpointer user_data);

typedef struct {
	FAMConnection  *fc;
	FAMRequest     *fr;
	GList 	       *fl;

	gchar          *path_name;
	gchar	       *monitor_name;

	FamAddedFunc       added_func;
	FamRemovedFunc     removed_func;

	GAsyncQueue    *queue;
	GThread	       *thread;

	gpointer       *user_data;
	GDestroyNotify  destroy;

	gboolean        terminate;
} FamMonitorWatch;

GType
fam_monitor_get_type (void);

FamMonitor*
fam_monitor_new (void);

/* Currently only for monitoring directories */
void
fam_monitor_watch_new (FamMonitor *self, 
		       gchar *monitor_name, 
                       gchar *path_name, 
                       FamAddedFunc added_func, 
                       FamRemovedFunc removed_func, 
                       gpointer user_data, 
                       GDestroyNotify destroy);

void
fam_monitor_watch_terminate (FamMonitor *monitor, gchar *monitor_name);

#endif /* FAM_MONITOR_H */
